// ViewAudio.dll v0.3.01 Copyright(C) 2003 minamina
// Avisynth Plugin - Cache Audio

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
// http://www.gnu.org/copyleft/gpl.html .

#include "CacheAudio.h"

CacheAudio::CacheAudio(PClip _child, int nMaxFrame, int nBlockFrame, IScriptEnvironment* env)
	: GenericVideoFilter(_child)
{
	if (!vi.HasAudio())
	{
		env->ThrowError("%s : clip has no audio", GetName());
	}

	if (nMaxFrame < 0)
	{
		nMaxFrame = vi.num_frames;
	}
	nMaxFrame = min(max(1, nMaxFrame), vi.num_frames);

	if (nBlockFrame < 0)
	{
		nBlockFrame = nMaxFrame;
	}
	nBlockFrame = min(max(1, nBlockFrame), nMaxFrame);

	m_nBytesPerAudioSample = vi.BytesPerAudioSample();

	m_nMaxBlock = nMaxFrame / nBlockFrame;
	m_nBlockSample = (int)vi.AudioSamplesFromFrames(nBlockFrame);
	m_nBlockSize = m_nBlockSample * m_nBytesPerAudioSample;
}

PVideoFrame __stdcall CacheAudio::GetFrame(int n, IScriptEnvironment *env)
{
	return child->GetFrame(n, env);
}

void __stdcall CacheAudio::GetAudio(void* buf, __int64 start, __int64 count, IScriptEnvironment* env)
{
	BYTE* pBuffer = (BYTE*)buf;
	BlockBuffer* pBlockBuffer;
	int end = (int)(start + count);

	for (int i = (int)start; i < end; i++)
	{
		if (i < 0 || i > vi.num_audio_samples)
		{
			::ZeroMemory(pBuffer, m_nBytesPerAudioSample);
		}
		else
		{
			pBlockBuffer = GetCacheBlockBuffer(i, env);
			::memcpy(pBuffer, pBlockBuffer->GetBuffer() + ((i % m_nBlockSample) * m_nBytesPerAudioSample), m_nBytesPerAudioSample);
		}
		pBuffer += m_nBytesPerAudioSample;
	}
}

CacheAudio::BlockBuffer* CacheAudio::GetCacheBlockBuffer(const int nStartSample, IScriptEnvironment* env)
{
	int nBlockStart = (nStartSample / m_nBlockSample) * m_nBlockSample;

	BlockBufferList::iterator it;
	for (it = m_listBlockBuffer.begin(); it != m_listBlockBuffer.end(); ++it)
	{
		if ((*it)->GetStartSample() == nBlockStart)
		{
			break;
		}
	}

	if (it != m_listBlockBuffer.end())
	{
		if (it != m_listBlockBuffer.begin())
		{
			m_listBlockBuffer.splice(m_listBlockBuffer.begin(), m_listBlockBuffer, it);
		}
	}
	else
	{
		if (m_listBlockBuffer.size() >= m_nMaxBlock)
		{
			--it;
			m_listBlockBuffer.splice(m_listBlockBuffer.begin(), m_listBlockBuffer, it);
		}
		else
		{
			(*it) = new BlockBuffer(m_nBlockSize);
			m_listBlockBuffer.push_front((*it));
		}

		(*it)->SetStartSample(nBlockStart);
		child->GetAudio((*it)->GetBuffer(), nBlockStart, m_nBlockSample, env);
	}

	return (*it);
}
