// ViewAudio.dll v0.3.01 Copyright(C) 2003 minamina
// Avisynth Plugin - Cache Audio

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
// http://www.gnu.org/copyleft/gpl.html .

#ifndef _CACHE_AUDIO_H
#define _CACHE_AUDIO_H

#ifdef _MSC_VER
	#include <crtdbg.h>
#else
	#define _ASSERTE(x) assert(x)
	#include <assert.h>
#endif

#include <windows.h>

#pragma warning (push, 3)
//#include <boost/shared_ptr.hpp>
//#include <boost/lexical_cast.hpp>
#include "avisynth.h"
#include <list>
#pragma warning (pop)

const int DEFAULT_MAX_FRAME = 1000;
const int DEFAULT_BLOCK_FRAME = 100;

class CacheAudio : public GenericVideoFilter
{
private:
	class BlockBuffer
	{
	public:
		BlockBuffer(const int nSize) { m_pBuffer = new BYTE[nSize]; m_nStartSample = -1; }
		virtual ~BlockBuffer() { if (m_pBuffer != NULL) { delete[] m_pBuffer; m_pBuffer = NULL; } }
		BYTE* GetBuffer() { return m_pBuffer; }
		void SetStartSample(const int nStartSample) { m_nStartSample = nStartSample; }
		int GetStartSample() { return m_nStartSample; }
	private:
		int m_nStartSample;
		BYTE* m_pBuffer;
	};

private:
	int m_nBytesPerAudioSample;
	int m_nBlockSample;
	unsigned int m_nMaxBlock;
	int m_nBlockSize;
	typedef std::list<BlockBuffer*> BlockBufferList;
	BlockBufferList m_listBlockBuffer;

public:
	static void PluginInit(IScriptEnvironment *env)
	{ env->AddFunction(GetName(), GetParams(), Create, 0); }

	CacheAudio(PClip _child, int nMaxFrame, int nBlockFrame, IScriptEnvironment* env);
	virtual ~CacheAudio()
	{
		for (BlockBufferList::const_iterator it = m_listBlockBuffer.begin();
			it != m_listBlockBuffer.end(); ++it)
		{
			delete (*it);
		}
		m_listBlockBuffer.clear();
	};

private:
	BlockBuffer* GetCacheBlockBuffer(const int nStartSample, IScriptEnvironment* env);
	static AVSValue __cdecl Create(AVSValue args, void* /*user_data*/, IScriptEnvironment *env)
	{ return new CacheAudio(args[0].AsClip(), args[1].AsInt(DEFAULT_MAX_FRAME), args[2].AsInt(DEFAULT_BLOCK_FRAME), env); }
	static const char *GetName() { return "CacheAudio"; }
	static const char *GetParams() { return "c[MaxFrame]i[BlockFrame]i"; }

	PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment *env);
	void __stdcall GetAudio(void* buf, __int64 start, __int64 count, IScriptEnvironment* env);
};

#endif // _CACHE_AUDIO_H
