// ViewAudio.dll v0.3.01 Copyright(C) 2003 minamina
// Avisynth Plugin - View Audio (YUY2 and YV12 Only)

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
// http://www.gnu.org/copyleft/gpl.html .

#include "ViewAudio.h"
#include "CacheAudio.h"

#define MAX_SIDE_FRAMES (10)

extern "C" __declspec(dllexport)
const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env)
{
	ViewAudio::PluginInit(env);
	CacheAudio::PluginInit(env);
	return 0;
}

ViewAudio::ViewAudio(PClip _child, const int nSideFrames, const int nDrawBoxTop, const int nDrawBoxHeight,
					 const bool bFillDrawBox, const int nDrawChannel, IScriptEnvironment* env)
	: GenericVideoFilter(_child), m_pAudioBuffer(NULL)
{
	if(!vi.IsYUY2() && !vi.IsYV12())
	{
		env->ThrowError("%s : YUY2, YV12 only", GetName());
	}
	if (!vi.HasAudio())
	{
		env->ThrowError("%s : clip has no audio", GetName());
	}

	m_nSideFrames = nSideFrames;
	if (m_nSideFrames < 0) { m_nSideFrames = 0; }
	if (m_nSideFrames > MAX_SIDE_FRAMES) { m_nSideFrames = MAX_SIDE_FRAMES; }

	m_nDrawFrames = m_nSideFrames * 2 + 1;

	m_nDrawBoxHeight = nDrawBoxHeight;
	if (m_nDrawBoxHeight < 10) { m_nDrawBoxHeight = DEFAULT_DRAW_BOX_HEIGHT; }
	if (m_nDrawBoxHeight > vi.height) { m_nDrawBoxHeight = vi.height; }

	m_nDrawBoxTop = nDrawBoxTop;
	if (m_nDrawBoxTop < -1) { m_nDrawBoxTop = -1; }
	if (m_nDrawBoxTop > vi.height - m_nDrawBoxHeight) { m_nDrawBoxTop = vi.height - m_nDrawBoxHeight; }

	m_bFillDrawBox = bFillDrawBox;
	if (m_nDrawBoxTop == -1)
	{
		m_bFillDrawBox= true;
		m_nDrawBoxTop = vi.height;
		vi.height += m_nDrawBoxHeight;
	}

	m_nAudioChannels = vi.AudioChannels();
	m_nDrawChannel = nDrawChannel;
	if (m_nDrawChannel < -1) { m_nDrawChannel = -1; }
	if (m_nDrawChannel >= m_nAudioChannels) { m_nDrawChannel = -1; }

	m_nDrawAudioSamples = (int)vi.AudioSamplesFromFrames(m_nDrawFrames);
	int nBytesPerAudioSample = vi.BytesPerAudioSample();

	m_pAudioBuffer = new BYTE[m_nDrawAudioSamples * nBytesPerAudioSample];
}

PVideoFrame __stdcall ViewAudio::GetFrame(int n, IScriptEnvironment *env)
{
	PVideoFrame frameSrc = child->GetFrame(n, env);
	PVideoFrame frameDest = env->NewVideoFrame(vi);

	CopyFrame(frameSrc, frameDest, env);

	if (m_bFillDrawBox)
	{
		FillDrawBox(frameDest);
	}

	DrawBoxLine(frameDest);

	int nStartSamples = (int)vi.AudioSamplesFromFrames(n - m_nSideFrames);
	GetAudio(m_pAudioBuffer, nStartSamples, m_nDrawAudioSamples, env);

	if (m_nDrawChannel == -1)
	{
		int nChannel;
		for (nChannel = 0; nChannel < m_nAudioChannels; nChannel++)
		{
			DrawAudioChannel(frameDest, nChannel);
		}
	}
	else
	{
		DrawAudioChannel(frameDest, m_nDrawChannel);
	}

	return frameDest;
}

inline void ViewAudio::CopyFrame(const PVideoFrame& frameSrc, const PVideoFrame& frameDest, IScriptEnvironment *env)
{
	const BYTE *pSrc = frameSrc->GetReadPtr();
	BYTE *pDest = frameDest->GetWritePtr();

	int nPitchSrc = frameSrc->GetPitch();
	int nPitchDest = frameDest->GetPitch();
	int nRowSizeDest = frameDest->GetRowSize();
	int nHeightSrc = frameSrc->GetHeight();

	env->BitBlt(pDest, nPitchDest, pSrc, nPitchSrc, nRowSizeDest, nHeightSrc);

	if (vi.IsYV12())
	{
		const BYTE *pSrcU = frameSrc->GetReadPtr(PLANAR_U);
		BYTE *pDestU = frameDest->GetWritePtr(PLANAR_U);
		const BYTE *pSrcV = frameSrc->GetReadPtr(PLANAR_V);
		BYTE *pDestV = frameDest->GetWritePtr(PLANAR_V);
		int nPitchSrcU = frameSrc->GetPitch(PLANAR_U);
		int nPitchDestU = frameDest->GetPitch(PLANAR_U);
		int nRowSizeDestU = frameDest->GetRowSize(PLANAR_U);
		int nHeightSrcU = frameSrc->GetHeight(PLANAR_U);

		env->BitBlt(pDestU, nPitchDestU, pSrcU, nPitchSrcU, nRowSizeDestU, nHeightSrcU);
		env->BitBlt(pDestV, nPitchDestU, pSrcV, nPitchSrcU, nRowSizeDestU, nHeightSrcU);
	}
}

inline void ViewAudio::FillDrawBox(const PVideoFrame& frameDest)
{
	BYTE *pDest = frameDest->GetWritePtr();
	int nPitchDest = frameDest->GetPitch();

	if (vi.IsYUY2())
	{
		int nPos;
		int nPosMax = m_nDrawBoxHeight * nPitchDest >> 2;
		DWORD* pBuffer = (DWORD*)&pDest[m_nDrawBoxTop * nPitchDest];
		for (nPos = 0; nPos < nPosMax; nPos++)
		{
			pBuffer[nPos] = 0x80208020;
		}
	}
	else
	{
		BYTE *pDestU = frameDest->GetWritePtr(PLANAR_U);
		BYTE *pDestV = frameDest->GetWritePtr(PLANAR_V);
		int nPitchDestU = frameDest->GetPitch(PLANAR_U);

		::FillMemory(&pDest[m_nDrawBoxTop * nPitchDest], m_nDrawBoxHeight * nPitchDest, 0x20);
		::FillMemory(&pDestU[(m_nDrawBoxTop >> 1) * nPitchDestU], (m_nDrawBoxHeight >> 1) * nPitchDestU, 0x80);
		::FillMemory(&pDestV[(m_nDrawBoxTop >> 1) * nPitchDestU], (m_nDrawBoxHeight >> 1) * nPitchDestU, 0x80);
	}
}

inline void ViewAudio::DrawBoxLine(const PVideoFrame& frameDest)
{
	BYTE *pDest = frameDest->GetWritePtr();
	int nPitchDest = frameDest->GetPitch();

	int nRowSizeDest = frameDest->GetRowSize();
	BYTE* pBuffer = &pDest[m_nDrawBoxTop * nPitchDest];
	int nStep = vi.IsYUY2() + 1;
	int nWidth = nRowSizeDest / nStep;
	int nLowLinePos = (m_nDrawBoxHeight - 1) * nPitchDest;
	int nX;
	for (nX = 0; nX < nWidth; nX++)
	{
		pBuffer[nX * nStep] = 255;
		pBuffer[nX * nStep + nLowLinePos] = 255;
	}

	int nBoxWidth = nWidth / m_nDrawFrames;
	int nBoxNumber, nY, nBoxPos;
	for (nBoxNumber = 1; nBoxNumber < m_nDrawFrames; nBoxNumber++)
	{
		nBoxPos = nBoxWidth * nBoxNumber * nStep;
		for (nY = 0; nY < 4; nY++)
		{
			pBuffer[nBoxPos + nPitchDest * nY] = 255;
			pBuffer[nBoxPos - nPitchDest * nY + nLowLinePos] = 255;
		}
	}
}

inline void ViewAudio::DrawAudioChannel(const PVideoFrame& frameDest, const int nChannel)
{
	BYTE *pDest = frameDest->GetWritePtr();

	int nPitchDest = frameDest->GetPitch();
	int nRowSizeDest = frameDest->GetRowSize();

	BYTE* pBuffer = &pDest[m_nDrawBoxTop * nPitchDest];
	int nStep = vi.IsYUY2() + 1;
	int nWidth = nRowSizeDest / nStep;
	int nX, nY;
	for (nX = 0; nX < nWidth; nX++)
	{
		int nPosAudioBuffer = m_nDrawAudioSamples * nX / nWidth * m_nAudioChannels + nChannel;
		BYTE byteY;
		if (vi.sample_type == SAMPLE_INT8)
		{
			byteY = m_pAudioBuffer[nPosAudioBuffer];
		}
		else
		{
			// SAMPLE_INT16
			byteY = (BYTE)((((WORD*)m_pAudioBuffer)[nPosAudioBuffer] + 32768 + 128) >> 8);
		}
		nY = byteY * m_nDrawBoxHeight >> 8;

//		pBuffer[nY * nPitchDest + nX * nStep] ^= 255;
		pBuffer[nY * nPitchDest + nX * nStep] = 255;
	}
}
